#include <string>
#include <vector>
#include <iostream>
#include <pthread.h>
#include <unistd.h>
#include <fstream>

#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>

#define STRING	std::string
#define CIN	std::cin
#define COUT	std::cout
#define ENDL	std::endl
#define VECTOR	std::vector<const char *>
#define SVECT	std::vector<STRING>
#define IVECT	std::vector<int>

#define UNALLOC	0
#define INMEM	1
#define ONDISK	2

#define	MYPORT	3131	// the port users will be connecting to
#define	BACKLOG	10		// how many pending connections queue will hold
#define FRAMNUM 10

pthread_mutex_t manager_mutex = PTHREAD_MUTEX_INITIALIZER;

STRING allcommands[] = { "create", "access", "terminate"};

std::ofstream logfile;

class Frame
{
public:
	Frame()
	{
		free = true;
		accessbit = false;
	}

	bool isFree()
	{
		return free;
	}

	void freeIt()
	{
		free = true;
		accessbit = false;
	}

	bool isAccessed()
	{
		if(free)
			return false;

		bool ret = accessbit;
		accessbit = false;
		return ret;
	}

	STRING getName()
	{		
		return process;
	}

	bool getAccess()
	{
		return accessbit;
	}

	void setAccess()
	{
		accessbit = true;
	}

	void allocate(STRING newprocess)
	{
		free = false;
		process = newprocess;
		accessbit = true;
	}
private:
	bool free;
	bool accessbit;
	STRING process;
};

class Page
{
public:
	Page()
	{
		status = UNALLOC;
	}
	
	int getStatus()
	{
		return status;
	}

	int getFrameNum()
	{
		return framenum;
	}

	void allocate(int framenumber)
	{
		framenum = framenumber;
		status = INMEM;
	}

	void swap()
	{
		status = ONDISK;
	}
private:
	int framenum;
	int status;
};

#define FRMVECT	std::vector<Frame>
#define PGVECT	std::vector<Page>

class PageTable
{
public:
	PageTable(STRING processname, int size)
	{
		name = processname;
		pages.resize(size);
	}

	bool compareName(STRING tocompare)
	{
		return (name == tocompare);
	}

	int getSize()
	{
		return pages.size();
	}

	Page & getPage(int index)
	{
		return pages[index];
	}

	STRING getName()
	{
		return name;
	}
	
	int swap(int framenum)
	{
		for(int i = 0; i < pages.size(); i++)
		{
			if(pages[i].getStatus() == INMEM && pages[i].getFrameNum() == framenum)
			{
				logfile << "Page number " << i << " of process " << name
					<< " has swapped from frame number " << framenum << ENDL;

				pages[i].swap();
				return i;
			}
		}

		return -1;
	}
private:
	PGVECT pages;
	STRING name;
};

#define PTVECT	std::vector<PageTable>

class Manager 
{
public:
	Manager(int i)
	{
		clock = hit = miss = access = 0;
		frames.resize(FRAMNUM);
		
		for(int i = FRAMNUM - 1; i >= 0; i--)
			freelist.push_back(i);
	}
	
	void hitRate()
	{
		COUT << "Hit Rate: ";
	
		if(access == 0)
			COUT << "100";
		else
			COUT << int((float(hit) / float(access)) * 100);

		COUT << " %" << ENDL;
	}
	
	void missRate()
	{
		COUT << "Miss Rate: ";
	
		if(access == 0)
			COUT << "100";
		else
			COUT << int((float(miss) / float(access)) * 100);

		COUT << " %" << ENDL;

	}

	void memStat()
	{
		COUT << "Mem Stat:" << ENDL;

		for(int i = 0; i < frames.size(); i++)
		{
			COUT << "\tFrame#" << i << ": ";

			if(frames[i].isFree())
				COUT << "Free";
			else
				COUT << frames[i].getName();

			if(frames[i].getAccess())
				COUT << " Accessed" << ENDL;
			else
				COUT << " Not accessed" << ENDL;
		}
	}

	void memUsage()
	{
		COUT << "Mem Usage:" << ENDL;
		COUT << "\tTotal frames: " << FRAMNUM << ENDL;
		COUT << "\tFree frames: " << freelist.size() << ENDL;
	}

	STRING deleteProcess(STRING name)
	{
		pthread_mutex_lock(&manager_mutex);

		logfile << "In deleteProcess:" << ENDL;
		
		int prcspt = findPageName(name);

		logfile << "The page found in index " << prcspt << ENDL;
		
		if(prcspt == -1)
		{
			logfile << "The process name for delete is not valid" << ENDL;
			pthread_mutex_unlock(&manager_mutex);
			return "Error: the process name for delete is not valid **ASSERT**";
		}

		int size = pagetables[prcspt].getSize();

		logfile << "The page found with size " << size << ENDL;

		for(int i = 0; i < size; i++)
		{
			if(pagetables[prcspt].getPage(i).getStatus() == INMEM)
			{
				int frmidx = pagetables[prcspt].getPage(i).getFrameNum();
				freelist.push_back(frmidx);
				frames[frmidx].freeIt();
			}
		}

		logfile << "The process " << pagetables[prcspt].getName() << "has terminated" << ENDL;

		pagetables.erase(PTVECT::iterator(&pagetables[prcspt]));

		pthread_mutex_unlock(&manager_mutex);

		return "The process has been terminated";
	}

	STRING accessPage(STRING name, int number)
	{
		pthread_mutex_lock(&manager_mutex);

		logfile << "In accessPage:" << ENDL;
		
		int prcspt = findPageName(name);

		if(prcspt == -1)
		{
			logfile << "The process name for access is not valid: " << name << ENDL;
			pthread_mutex_unlock(&manager_mutex);
			return "Error: the process name for access is not valid **ASSERT**";
		}

		int size = pagetables[prcspt].getSize();
		
		if(size <= number)
		{
			logfile << "Page address is invalid" << ENDL;
			pthread_mutex_unlock(&manager_mutex);
			return "Err: page address is invalid";
		}

		Page & thepage = pagetables[prcspt].getPage(number);
		STRING retval;
		
		if(thepage.getStatus() != INMEM)
		{
			if(thepage.getStatus() == ONDISK)
				retval = "The page swaped back into ";
			else
				retval = "The page allocated ";
			
			if(hasFree())
			{
				retval += "free frame# ";
				
				int myfreeframe = freelist[freelist.size() - 1];
				freelist.pop_back();
				frames[myfreeframe].allocate(name);
				
				pagetables[prcspt].getPage(number).allocate(myfreeframe);

				retval += int2str(myfreeframe);
			}
			else
			{
				retval += "frame# ";
				int myframenumber = findLRU();
				clock = (clock + 1) % FRAMNUM;

				int owner = findPageName(frames[myframenumber].getName());

				if(owner == -1)
				{
					logfile << "The owner does not exist!!" << ENDL;
					pthread_mutex_unlock(&manager_mutex);
					return "Error: The owner does not exist **ASSERT**";
				}
				
				int swaped = pagetables[owner].swap(myframenumber);

				if(swaped == -1)
				{
					logfile << "The owner has not the page!!" << ENDL;
					pthread_mutex_unlock(&manager_mutex);
					return "Error: The owner has not the page **ASSERT**";
				}
				
				frames[myframenumber].allocate(name);
				pagetables[prcspt].getPage(number).allocate(myframenumber);
				retval += int2str(myframenumber);
				retval += " instead of page# ";
				retval += int2str(swaped);
				retval += " from process ";
				retval += pagetables[owner].getName();
			}

			access++;
			miss++;
		}
		else
		{
			access++;
			hit++;
			retval = "The process accessed in frame# ";
			retval += int2str(thepage.getFrameNum());
			retval += " in memmory";
			frames[thepage.getFrameNum()].setAccess();
		}

		pthread_mutex_unlock(&manager_mutex);
	
		return retval;
	}

	STRING createProcess(STRING name, int number)
	{
		pthread_mutex_lock(&manager_mutex);

		logfile << "In createProcess:" << ENDL;
		
		int prcspt = findPageName(name);

		if(prcspt != -1)
		{
			pthread_mutex_unlock(&manager_mutex);
			logfile << "The process name " << name << " is already exist" << ENDL;
			return "Error: the process name for is already exist";
		}

		pagetables.push_back(PageTable(name, number));
	
		pthread_mutex_unlock(&manager_mutex);

		return "The process has been created";
	}

	void dump()
	{
		logfile << "DUMP" << ENDL;
		logfile << "{" << ENDL;
		
		for(int i = 0; i < pagetables.size(); i++)
			logfile << '\t' << i << ": " << pagetables[i].getName() << ENDL;

		logfile << "}" << ENDL;
	}

private:
	STRING int2str(int num)
	{
		STRING temp;

		if(num <= 0)
			return "0";

		for(;num != 0; num /= 10)
			temp += '0' + num - ((num / 10) * 10);

		STRING retval;

		for(int i = temp.size() - 1; i >= 0; i--)
			retval += temp[i];

		return retval;
	}

	int findPageName(STRING name)
	{
		for(int i = 0; i < pagetables.size(); i++)
		{
			if(pagetables[i].compareName(name))
				return i;
		}

		return -1;
	}

	bool hasFree()
	{
		if(freelist.size() == 0)
			return false;

		return true;
	}

	int findLRU()
	{
		while(frames[clock].isAccessed())
			clock = (clock + 1) % FRAMNUM;

		return clock;
	}
	
	int hit;
	int miss;
	int access;
	
	IVECT freelist;
	int clock;
	
	FRMVECT frames;
	PTVECT pagetables;
};

Manager manager(1);

void *service_requests(void *parameter)
{
	char temp[256];
	logfile << "In service_requests:" << ENDL;
	int new_fd = int(parameter);
	SVECT task;
	int thesize;
	int iteration;
	STRING msg;
	
	if(recv(new_fd, &iteration, sizeof(int), 0) != -1)
	{
		logfile << "Receiving:" << ENDL << "{" << ENDL;
		logfile << '\t' << "Number of elements: " << iteration << ENDL;
		
		task.resize(iteration);
	
		for(int i = 0; i < iteration; i++)
		{
			recv(new_fd, &thesize, sizeof(int), 0);
			logfile << '\t' << "Size of the element: " << thesize << ENDL;
			recv(new_fd, temp, thesize, 0);
			task[i] = temp;
			task[i].resize(thesize);
			logfile << '\t' << "Elements: " << task[i] << ENDL;	
		}

		logfile << "}" << ENDL;
		logfile << task[0] << "Received." << ENDL;

		if(task[0] == allcommands[0])
		{
			int temp;
			
			if(sscanf(task[2].c_str(),"%d", &temp) != 1)
				msg = "The third parameter must be an integer";
			else
				msg = manager.createProcess(task[1], temp);
		}
		else if(task[0] == allcommands[1])
		{
			int temp;
			
			if(sscanf(task[2].c_str(),"%d", &temp) != 1)
				msg = "The third parameter must be an integer";
			else
				msg = manager.accessPage(task[1], temp);
		}
		else if(task[0] == allcommands[2])
		{
			msg = manager.deleteProcess(task[1]);
		}
		else
		{
			msg = "ME: Unknown command!!";
			logfile << "It's been checked why error!!" << ENDL;
		}

		send(new_fd, msg.c_str(), 255, 0);
	}
	else
		logfile << "Cannot receive info!!" << ENDL;
			
	shutdown(new_fd, SHUT_RDWR);
	close(new_fd);

	return NULL;
}

void *listener(void *parameter)
{
	logfile << "In listener:" << ENDL;
	int sockfd = *(int *)parameter;
	int new_fd; // new connection on new_fd
	struct sockaddr_in my_addr; // my address information
	struct sockaddr_in their_addr; // connectors address information
	socklen_t sin_size;
	pthread_t thread_id;
	
	my_addr.sin_family = AF_INET; // host byte order
	my_addr.sin_port = htons(MYPORT); // short, network byte order
	my_addr.sin_addr.s_addr = INADDR_ANY; // auto-fill with my IP
	memset(&(my_addr.sin_zero), '\0', 8); // zero the rest of the struct

	if(bind(sockfd, (struct sockaddr *)&my_addr, sizeof(struct sockaddr)) == -1)
	{
		logfile << "Cannot bind address to socket" << ENDL;
		return NULL;
	}

	while(1)
	{
		if(listen(sockfd, BACKLOG) == -1)
		{
			logfile << "Cannot listen to socket" << ENDL;
			break;
		}

		sin_size = sizeof(struct sockaddr_in);
		new_fd = accept(sockfd, (struct sockaddr *)&their_addr, &sin_size);

		if(new_fd == -1)
		{
			logfile << "Error in accepting new socket" << ENDL;
			break;
		}
		
		pthread_create(&thread_id, NULL, &service_requests, (void *)new_fd);
	}

	return NULL;
}

int main()
{
	int args;
	char command[256];
	STRING allcommands[] = { "hitrate", "missrate", "memstat", "memusage", "terminate", "dump"};
	char * argv[256];
	command[0] = 0;
	logfile.open("server.log");

	int sockfd; // listen on sock_fd, new connection on new_fd
	sockfd = socket(AF_INET, SOCK_STREAM, 0);

	if (sockfd == -1)
	{
		logfile << "Cannot stablish port" << ENDL;
		logfile.close();
		return -1;
	}

	pthread_t thread_id;
	pthread_create(&thread_id, NULL, &listener, &sockfd);

	for(;;)
	{
		COUT << "/> ";
		CIN.getline( command, 255);
		
		while(command[0] == 0)
			CIN.getline( command, 255);
		
		args = 0;

		for( int ind = 0; command[ind] != 0;)
		{
			argv[args] = command + ind;			
				
			while( command[ind] != ' ' && command[ind] != 0)
				ind++;

			if( command[ind] != 0)
			{
				command[ind] = 0;
				ind++;
			}

			args++;	
			argv[args] = NULL;
		}
	
		if( argv[0] == NULL)
			continue;

		if( allcommands[0] == argv[0])
		{
			if( args == 1)
			{
				manager.hitRate();
			}
			else
				COUT << "The hitrate instruction gets 0 parameters !!" << ENDL;
		}
		else if( allcommands[1] == argv[0])
		{
			if( args == 1)
			{
				manager.missRate();
			}
			else
				COUT << "The missrate instruction gets 0 parameters !!" << ENDL;
		}
		else if( allcommands[2] == argv[0])
		{	
			if( args == 1)
			{
				manager.memStat();
			}
			else				
				COUT << "The memstat instruction gets 0 parameters !!" << ENDL;		
		}
		else if( allcommands[3] == argv[0])
		{
			if( args == 1)
			{
				manager.memUsage();
			}
			else				
				COUT << "The memusage instruction gets 0 parameters !!" << ENDL;
		}
		else if( allcommands[4] == argv[0])
		{
			if( args == 1)
			{
				break;
			}
			else				
				COUT << "The terminate instruction gets 0 parameters !!" << ENDL;
		}
		else if( allcommands[5] == argv[0])
		{
			if( args == 1)
			{
				manager.dump();
			}
			else				
				COUT << "The terminate instruction gets 0 parameters !!" << ENDL;
		}
		else
		{
			COUT << "Unknown command !!" << ENDL;
			continue;
		}
	}

	shutdown(sockfd, SHUT_RDWR);
	close(sockfd);

	logfile.close();
	return 0;
}


